<?php
App::uses('EvShopAppModel', 'EvShop.Model');
App::uses('ArrayUtil', 'EvCore.Lib');
App::uses('CakeNumber', 'Utility');

/**
 * Variant Model
 *
 * @property Product $Product
 * @property Option $Option
 */
class Variant extends EvShopAppModel {

	public $displayName = 'Pricing';

	/**
	 * Display field
	 *
	 * @var string
	 */
	public $displayField = 'name';

	/**
	 * Validation rules
	 *
	 * @var array
	 */
	public $validate = array(
		'product_id' => array(
			'numeric' => array(
				'rule' => array('numeric')
			),
		),
		'name' => array(
			'notEmpty' => array(
				'rule' => array('notBlank'),
				'message' => 'A variant name must be entered.'
			),
		),
		'price' => array(
			'notEmpty' => array(
				'rule' => array('notBlank')
			),
			'money' => array(
				'rule' => array('money')
			),
		),
	);

	//The Associations below have been created with all possible keys, those that are not needed can be removed

	/**
	 * belongsTo associations
	 *
	 * @var array
	 */
	public $belongsTo = array(
		'Product' => array(
			'className' => 'EvShop.Product'
		)
	);

	/**
	 * hasAndBelongsToMany associations
	 *
	 * @var array
	 */
	public $hasAndBelongsToMany = array(
		'Option' => array(
			'className' => 'EvShop.Option',
			'joinTable' => 'ev_shop_options_variants'
		)
	);

	public $findMethods = array(
		'allAndRound' => true
	);

	/**
	 * get all the data and then round the values
	 *
	 * @param 	string 	state - either before or after
	 * @param 	array 	array of the query params
	 * @param 	array 	Array of results (after only)
	 * @return 	array 	$query if state is before / $results if state is after
	 */
	protected function _findAllAndRound($state, $query, $results = array()) {
		if ($state === 'after') {
			foreach ($results as $key => $result) {
				$results[$key]['Variant']['rrp'] = CakeNumber::format($result['Variant']['rrp'], 2);
				$results[$key]['Variant']['price'] = CakeNumber::format($result['Variant']['price'], 2);
				$results[$key]['Variant']['sale_price'] = CakeNumber::format($result['Variant']['sale_price'], 2);

				$results[$key]['optionDesc'] = Hash::combine(
					$result,
					'Option.{n}.OptionGroup.name',
					'Option.{n}.name'
				);
				array_walk(
					$results[$key]['optionDesc'],
					function (&$value, $key) {
						$value = $key . ': ' . $value;
					}
				);
				$results[$key]['optionDesc'] = implode(', ', $results[$key]['optionDesc']);
			}

			return $results;
		}

		return $query;
	}

	/**
	 * get listing for variants management
	 *
	 * @param 	int 	Product Id
	 * @param 	array 	custom array conditions if needed
	 * @return 	array 	results
	 */
	public function manageListing($productId, $query = array()) {
		$params = array(
			'conditions' => array(
				'Variant.product_id' => $productId
			),
			'order' => 'Variant.sequence ASC, Variant.id ASC',
			'contain' => array(
				'Option' => array(
					'OptionGroup'
				)
			)
		) + $query;

		return $this->find(
			'allAndRound',
			$params
		);
	}

	/**
	 * afterSave - send out variant saved event
	 *
	 */
	public function afterSave($created, $options = array()) {
		parent::afterSave($created, $options);

		// dispatch product saved event
		$this->getEventManager()->dispatch(
			new CakeEvent('EvShop.Model.Variant.saved', $this, array(
				'newItem' => $created
			))
		);
	}

	/**
	 * get any changes in the options and return the new variants
	 *
	 * @param 	array 	Current Options
	 * @param 	array 	New Options
	 * @param 	string 	default product name
	 * @return 	array 	array of variants ready for DB insert
	 */
	public function processNewOptions($currentOptions, $newOptions, $productName) {
		if (empty($currentOptions) && empty($newOptions)) {
			return $this->defaultInsertArray($productName);
		}

		$options = $this->getNewOptions(
			array(
				'CurrentOptions' => $currentOptions,
				'Options' => $newOptions
			)
		);

		$variants = array();

		if (! empty($options)) {
			foreach ($options as $optionIds) {
				$variants += $this->buildVariants(
					$optionIds,
					$productName
				);
			}
		}

		return $variants;
	}

	/**
	 * calculate any new options
	 *
	 * @param 	array 	Array of Options / CurrentOptions elements
	 * @return 	array 	Options to build variants for
	 */
	public function getNewOptions($data) {
		// check if the main option groups match
		$currentOptions = array_keys($data['CurrentOptions']);
		sort($currentOptions);
		$options = array_keys($data['Options']);
		sort($options);

		// rebuild all the selected options if they don't match
		if ($currentOptions !== $options) {
			return array($data['Options']);
		}

		$returnOptions = array();

		// the option groups are the same, check which options have been added
		// i.e. adding more colours
		foreach ($data['Options'] as $groupId => $optionIds) {
			foreach ($optionIds as $id => $value) {

				if (! isset($data['CurrentOptions'][$groupId][$id])) {
					$returnOptions[$groupId][$groupId][$id] = $id;
				}
			}
		}

		// add to the returning optins array
		// all the other groups that haven't changed so we can build all the new variant combinations
		if (! empty($returnOptions)) {
			foreach ($returnOptions as $section => $group) {
				$options = $data['Options'];

				unset($options[key($group)]);
				$returnOptions[$section] += $options;
			}
		}

		return $returnOptions;
	}

	/**
	 * take the multiple option arrays and build variants
	 *
	 * @param 	array 	array of option ids
	 * @param 	string 	The entered product name as a fallback if no options were selected
	 * @return 	array 	array of variants ready for DB insertion
	 */
	public function buildVariants($options, $productName) {
		if (empty($options) || ! is_array($options)) {
			return array();
		}

		$Option = EvClassRegistry::init('EvShop.Option');
		$optionNames = $Option->getOptionNames(
			Hash::flatten($options)
		);

		$cartesian = ArrayUtil::cartesianProduct($options);

		return $this->generateInsertArray(
			$cartesian,
			$optionNames
		);
	}

	/**
	 * from the cartesian array generate db insert array
	 *
	 * @param 	array 	Cartesian product arrays
	 * @return 	array 	Array with variant name and option links
	 */
	public function generateInsertArray($cartesianProduct, $optionNames) {
		$variants = array();

		foreach ($cartesianProduct as $optionIds) {
			$name = array_intersect_key(
				$optionNames,
				array_flip($optionIds)
			);

			$variants[] = array(
				'name' => implode(' ', $name),
				'is_active' => 1,
				'Option' => array(
					'Option' => $optionIds
				)
			);
		}

		return $variants;
	}

	/**
	 * generate a default Insert Array for a product with no options
	 * variant = product name
	 *
	 * @param 	string 	Product name
	 * @return 	array 	Array with variant data for insert
	 */
	public function defaultInsertArray($productName) {
		return array(
			array(
				'name' => $productName,
				'is_active' => 1,
			)
		);
	}

	/**
	 * calculate the changes to options that require variants to be deleted
	 *
	 * @param 	array 	Array of Options / CurrentOptions elements
	 * @return 	array 	Options to remove
	 */
	public function getDeletedOptions($data) {
		// check if the main option groups match
		$currentOptions = array_keys($data['CurrentOptions']);
		sort($currentOptions);
		$options = array_keys($data['Options']);
		sort($options);

		// flag for deletion if not
		if ($currentOptions !== $options) {
			return Hash::flatten($data['CurrentOptions']);
		}

		// the option groups are the same, check which options have been removed
		$deletedOptions = array();

		foreach ($data['CurrentOptions'] as $groupId => $optionIds) {
			foreach ($optionIds as $id => $value) {

				if (! isset($data['Options'][$groupId][$id])) {
					$deletedOptions[] = $id;
				}
			}
		}

		return $deletedOptions;
	}

	/**
	 * process the deleted options and get the variants to delete
	 *
	 * @param 	array 	Current Options
	 * @param 	array 	New Options
	 * @param 	int 	product Id
	 * @return 	array 	Array of all the variants to delete
	 */
	public function processDeletedOptions($currentOptions, $newOptions, $productId) {
		$options = $this->getDeletedOptions(
			array(
				'CurrentOptions' => $currentOptions,
				'Options' => $newOptions
			)
		);

		if (empty($options)) {
			return null;
		}

		return $this->find(
			'list',
			array(
				'joins' => array(
					array(
						'table' => 'ev_shop_options_variants',
						'alias' => 'OptionsVariant',
						'conditions' => array(
							'Variant.id = OptionsVariant.variant_id'
						)
					)
				),
				'conditions' => array(
					'Variant.product_id' => $productId,
					'OptionsVariant.option_id' => $options
				)
			),
			array('callbacks' => false)
		);
	}

}
